#include <QGLWidget>
#include <QMessageBox>
#include <QImage>
#include <math.h>
#include "qtglutils.h"

/*
 *  Fabrication d'une DisplayListe openGL d'affichage d'une sphère
 *  Entrée :
 *     slices : Nombre de côtes d'orange de la planète
 *     rayon : rayon de la sphère
 *  Retour : numéro de la DisplayList générée.
 *
*/
GLuint newDlSphere(float rayon, int slices)
{
    GLuint DL_Sphere;
    GLUquadricObj *gBoule = gluNewQuadric();
    gluQuadricNormals(gBoule, GLU_TRUE); // Pour l'éclairage
    DL_Sphere = glGenLists(1);
    glNewList(DL_Sphere, GL_COMPILE);
    gluSphere(gBoule, rayon, slices, slices);
    glEndList();
    gluDeleteQuadric(gBoule); // Ne pas oublier
    return DL_Sphere;
}

/* Chargement d'une texture depuis un fichier image
 * Entrée :
 *      filename  : fichier image à charger
 *      minfilter : Minification filter
 *      magfilter : Magnification filter
 *      wrap      : Repetition ou clamp
 * Sortie :
 *      GLuint de l'objet texture ou 0 si erreur
*/

GLuint loadTexture(char *filename, GLenum minfilter, GLenum magfilter, GLenum wrap)
{
    GLenum      type;            /* Texture type */
    GLuint      textobj = 0;         /* Texture object */

    /* Try loading the bitmap file... */
    QImage image(QString(filename),NULL);
    if (image.isNull()) {
        QMessageBox::information(NULL,"ERREUR","Echec chargement image texture");
        return 0;
    }    
    QImage tex = QGLWidget::convertToGLFormat(image);

    if (image.height() == 1) type = GL_TEXTURE_1D;
    else type = GL_TEXTURE_2D;

    /* Create and bind a texture object */
    glGenTextures(1, &textobj);
    glBindTexture(type, textobj);
    /* Set texture parameters */
    glTexParameteri(type, GL_TEXTURE_MAG_FILTER, magfilter);
    glTexParameteri(type, GL_TEXTURE_MIN_FILTER, minfilter);
    glTexParameteri(type, GL_TEXTURE_WRAP_S, wrap);
    glTexParameteri(type, GL_TEXTURE_WRAP_T, wrap);
    glTexEnvi(type, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    if (minfilter == GL_LINEAR || minfilter == GL_NEAREST)
        glTexImage2D(type, 0, GL_RGBA, image.width(), image.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, tex.bits());
    else if (type == GL_TEXTURE_1D)
        gluBuild1DMipmaps(type, GL_RGBA, image.width(), GL_RGBA, GL_UNSIGNED_BYTE, tex.bits());
    else
        gluBuild2DMipmaps(type, GL_RGBA, image.width(), image.height(), GL_RGBA, GL_UNSIGNED_BYTE, tex.bits());

    return (textobj);
}

/*
 // *  Fabrication d'une DisplayListe openGL d'affichage d'une planète
 *  Entrée :
 *     rayon : rayon de la planète
 *     slices : Nombre de côtes d'orange de la planète
 *     nomfich : Nom du fichier où se trouve le bitmap du planisphère
 *  Retour : numéro de la DisplayList générée.
 *
*/
GLuint newDlPanete(float rayon, int slices, char *nomfich)
{
    GLuint dl_planet;
    //    GLuint theTexture = TextureLoad (nomfich, GL_FALSE, GL_NEAREST, GL_NEAREST, GL_REPEAT);
    GLuint theTexture = loadTexture (nomfich, GL_NEAREST, GL_NEAREST, GL_REPEAT);
    if (theTexture == 0) return 0;

    /* Le planisphere est colle sur la sphere autour de son axe Z. Les bords droite et gauche se rejoignent
       sur le premier slice qui est du coté +Y du repère de définition des points de la sphère.
       Il en résulte que le meridien de Greenwich qui est la verticale centrale du planisphere se retrouve
       du cote -Y. Avant d'appeler la définition des points et de la texture, on fait tourner le repère de
       90° autour de Z amenant -Y selon X, ce qui fait que l'axe X va passer par Greenwich. */
    dl_planet = glGenLists(1);                      // Nom de l'objet que l'on fabrique
    glShadeModel(GL_SMOOTH);                    // Pour que ce soit plus joli
    glNewList(dl_planet, GL_COMPILE);           // Ouverture de la définition de l'objet avec simple compilation
    glRotated(90.,0.,0.,1.);                    // On met l'axe des X dans le méridien de Greenwich
    GLUquadricObj *quadObj = gluNewQuadric();   // On va utiliser une structure de forme quadrique
    gluQuadricNormals(quadObj, GLU_TRUE);       // On génère les normales, nécessaire car couleurs fonction eclairage
    glBindTexture(GL_TEXTURE_2D, theTexture);   // On definit une texture à lier
    gluQuadricTexture(quadObj, GLU_TRUE);       // On asssocier la texture à la quadrique AVANT (important) sa creation
    glEnable(GL_TEXTURE_2D);                    // On autorise l'association d'une texture dans sa définition
    gluSphere(quadObj, rayon, slices, slices);  // Creation de la quadrique sphere
    glDisable(GL_TEXTURE_2D);                   // Important pour les autres objets
    glEndList();                                    // Fin de définition de l'objet
    gluDeleteQuadric(quadObj);                      // On n'a plus besoin de la structure de forme quadrique
    return dl_planet;
}
