#include <GL/glu.h>
#include <QAction>
#include "sacso.h"
#include <math.h>
#ifndef M_PI
#define M_PI   3.14159265358979323846
#endif
#define RAD2DEG (180./M_PI)
#define DEG2RAD (M_PI/180.)

#include "mllglutil.h"

/*---------------------------------------------------------------------
        Recopie d'un vecteur de R3 : U = V
        ----------------------------------

        Entrees:
                double *u	: Adresse du v3 affecté
                double *v	: Adresse vecteur donné

        Sorties:
                u est affecté
                Retour		: Adresse du v3 affecté
----------------------------------------------------------------------*/
static double *cpyv3(double* u, double* v)
{
    u[0] = v[0] ;
    u[1] = v[1] ;
    u[2] = v[2] ;
    return u;
}

/*---------------------------------------------------------------------
        Addition de vecteurs de R3 : U = V + W
        --------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du premier vecteur de R3 donné
                double *W	: Adresse du deuxième vecteur de R3 donné

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V ou W
----------------------------------------------------------------------*/
static double *addv3(double* u, double* v, double* w)
{
    u[0] = v[0] + w[0] ;
    u[1] = v[1] + w[1] ;
    u[2] = v[2] + w[2] ;
    return u;
}

/*---------------------------------------------------------------------
        Soustraction de vecteurs de R3 : U = V - W
        ------------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du premier vecteur de R3 donné
                double *W	: Adresse du deuxième vecteur de R3 donné

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V ou W
----------------------------------------------------------------------*/
static double *subv3(double* u, double* v, double* w)
{
    u[0] = v[0] - w[0] ;
    u[1] = v[1] - w[1] ;
    u[2] = v[2] - w[2] ;
    return u;
}
/*---------------------------------------------------------------------
        Carré d'un vecteurs de R3 : s = A'*A
        ------------------------------------

        Entrees:
                double *A	: Adresse vecteur donné

        Sorties:
                Retour		: Somme des carrés des composantes
----------------------------------------------------------------------*/
static double carrev3(double* a)
{
    return(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]);
}
/*---------------------------------------------------------------------
        Produit scalaire de vecteurs de R3 : s = <A . B>
        ------------------------------------------------

        Entrees:
                double *A	: Adresse du premier vecteur de R3 donné
                double *B	: Adresse du deuxième vecteur de R3 donné

        Sorties:
                Retour		: produit scalaire
----------------------------------------------------------------------*/
static double dotv3(double* a, double* b)
{
    return(a[0]*b[0]+a[1]*b[1]+a[2]*b[2]);
}
/*---------------------------------------------------------------------
        Longueur d'un vecteurs de R3 : s = sqrt(A'*A)
        ---------------------------------------------

        Entrees:
                double *A	: Adresse vecteur donné

        Sorties:
                Retour		: Longueur du vecteur
----------------------------------------------------------------------*/
static double longv3(double* a)
{
    return(sqrt(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]));
}
/*---------------------------------------------------------------------
        Normalisation d'un vecteur de R3 : A = B / ||B||
        ------------------------------------------------

        Entrees:
                double *A	: Adresse vecteur unitaire résultat
                double *B	: Adresse vecteur donné

        Sorties:
                A est affecté
                Retour		: Longueur du vecteur B

        Remarque : A et B peuvent occuper la même mémoire
                       si ||B|| = 0  -> A[0] = 1, les autres nuls.

        Remarque : A peut occuper la même mémoire que B
----------------------------------------------------------------------*/
static double lunitv3(double* a, double* b)
{
    double l ;
    l = sqrt(b[0]*b[0]+b[1]*b[1]+b[2]*b[2]);
    if (l == 0.)
    {
        a[0]=1.; a[1]=0.; a[2]=0.;
    }
    else
    {
        a[0] = b[0]/l;
        a[1] = b[1]/l;
        a[2] = b[2]/l;
    }
    return(l);
}
/*---------------------------------------------------------------------
        Multiplication d'un vecteur de R3 par un scalaire : U = sV
        ----------------------------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du vecteur de R3 donné
                double s	: Scalaire donné

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V
----------------------------------------------------------------------*/
static double *scav3(double* u, double s, double* v)
{
    u[0] = s * v[0] ;
    u[1] = s * v[1] ;
    u[2] = s * v[2] ;
    return u;
}
/*---------------------------------------------------------------------
        Produit vectoriel de vecteurs de R3 : A = B ^ C
        -----------------------------------------------

        Entrees:
                double *A	: Adresse du vecteur affecté
                double *B	: Adresse du premier vecteur de R3 donné
                double *C	: Adresse du deuxième vecteur de R3 donné

        Sorties:
                A est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : A peut occuper la même mémoire que B ou C
----------------------------------------------------------------------*/
static double *crossv3(double* a, double* b, double* c)
{
    double p, q ;
    p  = b[1] * c[2] - b[2] * c[1];
    q  = b[2] * c[0] - b[0] * c[2];
    a[2] = b[0] * c[1] - b[1] * c[0];
    a[0] = p ;
    a[1] = q ;
    return a;
}
/*---------------------------------------------------------------------
        Cumulation de vecteurs de R3 : U = V + aW
        -----------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du premier vecteur de R3 donné
                double a	: Scalaire donné
                double *W	: Adresse du deuxième vecteur de R3 donné

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V ou W
----------------------------------------------------------------------*/
static double *cumv3(double* u, double* v,double a, double* w)
{
    u[0] = v[0] + a*w[0] ;
    u[1] = v[1] + a*w[1] ;
    u[2] = v[2] + a*w[2] ;
    return u;
}
/*---------------------------------------------------------------------
        Produit matrice x vecteurs de R3 : w = A v
        -----------------------------------------------

        Entrees:
                double *w	: Adresse du vecteur affecté
                double *A	: Adresse de la matrice de R3 donné
                double *v	: Adresse du vecteur de R3 donné

        Sorties:
                w est affecté
                Retour		: Adresse du vecteur affecté

    Remarque : w et v peuvent occuper la même mémoire.
----------------------------------------------------------------------*/
static double *m3v3(double* w, double* a, double* v)
{
    double p, q ;
    p  = a[0] * v[0] + a[3] * v[1] + a[6] * v[2];
    q  = a[1] * v[0] + a[4] * v[1] + a[7] * v[2];
    w[2] = a[2] * v[0] + a[5] * v[1] + a[8] * v[2];
    w[0] = p ;
    w[1] = q ;
    return w;
}
static double *carol2mrot(double *a, double *carol) /* FROM rotation.c */
{
double cc, sc, ca, sa, cr, sr ;
double sasr, sacr ;

    cc = cos (carol[0]) ; ca = cos (carol[1]) ; cr = cos (carol[2]) ;
    sc = sin (carol[0]) ; sa = sin (carol[1]) ; sr = sin (carol[2]) ;
    sasr = sa * sr ; sacr = sa * cr ;

    a[0] = cc*ca  ; a[3] = cc*sasr - sc*cr ; a[6] = cc*sacr + sc*sr ;
    a[1] = sc*ca  ; a[4] = sc*sasr + cc*cr ; a[7] = sc*sacr - cc*sr ;
    a[2] =   -sa  ; a[5] =    ca*sr        ; a[8] =    ca*cr        ;
        return a;
}
/*---------------------------------------------------------------------
        Recopie d'un vecteur de Rn : U = V
        ----------------------------------

        Entrees:
                double *u	: Adresse du vecteur affecté
                double *v	: Adresse vecteur donné
                int n		: dimension de v

        Sorties:
                u est affecté
                Retour		: Adresse du vecteur affecté
----------------------------------------------------------------------*/
static double *cpyvec (double* u, double* v, int n)
{
    if (n > 0) while (n--) u[n] = v[n];
    return u;
}
/*---------------------------------------------------------------------
        Addition de vecteurs de Rn : U = V + W
        --------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du premier vecteur de Rn donné
                double *W	: Adresse du deuxième vecteur de Rn donné
                int n		: dimension des vecteurs

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V ou W
----------------------------------------------------------------------*/
static double *addvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] + w[n];
    return u;
}
/*---------------------------------------------------------------------
        Soustraction de vecteurs de Rn : U = V - W
        ------------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du premier vecteur de Rn donné
                double *W	: Adresse du deuxième vecteur de Rn donné
                int n		: dimension des vecteurs

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V ou W
----------------------------------------------------------------------*/
static double *subvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] - w[n];
    return u;
}

/*---------------------------------------------------------------------
        Multiplication d'un vecteur de Rn par un scalaire : U = aV
        ----------------------------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double a	: Scalaire donné
                double *V	: Adresse du vecteur de Rn donné
                int n		: dimension du vecteur

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V
----------------------------------------------------------------------*/
static double *scavec (double *u, double a, double *v, int n)
{
    if (n > 0) while (n--) u[n] = a * v[n] ;
    return u;
}
/*---------------------------------------------------------------------
        Addition de vecteurs de Rn : U = V + W
        --------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du premier vecteur de Rn donné
                double s	: Scalaire donné
                double *W	: Adresse du deuxième vecteur de Rn donné
                int n		: dimension des vecteurs

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V ou W
----------------------------------------------------------------------*/
static double *cumvec (double* u, double* v, double s, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] + s*w[n];
    return u;
}

/*---------------------------------------------------------------------
        Multiplication terme à terme de vecteurs de Rn : U = V * W
        ----------------------------------------------------------

        Entrees:
                double *U	: Adresse du vecteur affecté
                double *V	: Adresse du premier vecteur de Rn donné
                double *W	: Adresse du deuxième vecteur de Rn donné
                int n		: dimension des vecteurs

        Sorties:
                U est affecté
                Retour		: Adresse du vecteur affecté

        Remarque : U peut occuper la même mémoire que V ou W
----------------------------------------------------------------------*/
static double *mulvec (double* u, double* v, double* w, int n)
{
    if (n > 0) while (n--) u[n] = v[n] * w[n];
    return u;
}
/*---------------------------------------------------------------------
        Produit de matrice : Amp = Bmn x Cnp
        ------------------------------------

        Entrees:
                double *A	: Adresse de A
                double *B	: Adresse de B
                double *C	: Adresse de C
                int    m	: Nombre de lignes de A et de B
                int    n	: Nombre de colonnes de B et de lignes de C
                int    p	: Nombre de colonnes de A et de C

    Sortie :
                A est calculée
                Retour	: Adresse de A			:

    Remarque : première et dernière dimension : celles du résultat
        A ne doit pas occuper la même zone mémoire que B ou C
----------------------------------------------------------------------*/
static double *matmat (double *A, double *B, double *C, int m, int n, int p)
{
    double *a, *b, *c, r ;
    int i, j, k ;


    for (i = 0 ; i < m ; ++i)
    {
        a = A + i ;      			/* A(i,0) */
        c = C ;                                 /* C(0,0) */
        for (j = 0 ; j < p ; ++j)
        {
            r = 0 ;
            b = B + i ;				/* B(i,0) */
            for (k = 0 ; k < n ; ++k)
            {
                r += *b * *c++ ;		/* B(i,k) * C(k,j) */
                b += m ;
            }
            *a = r ;
            a += m ;				/* A(i,j) */
        }
    }
    return A;
}
/*---------------------------------------------------------------------
        Produit de matrice : Amp = [B']mn x Cnp
        ---------------------------------------

        Entrees:
                double *A	: Adresse de A
                double *B	: Adresse de B
                double *C	: Adresse de C
                int    m         : Nombre de lignes de A et de colonnes de B
                int    n         : Nombre de lignes de B et  de C
                int    p         : Nombre de colonnes de A et de C

    Sortie :
                A est calculée
                Retour	: Adresse de A			:

    Remarque : première et dernière dimension : celles du résultat
        A ne doit pas occuper la même zone mémoire que B ou C
----------------------------------------------------------------------*/
static double *tmatmat (double *a, double *b, double *c, int m, int n, int p)
{
    int i, j, k, ij, ki, kj, kj0 ;

    for (j = 0, ij = 0, kj0 = 0 ; j < p ; j++, kj0 += n)
    {
        for (i = 0, ki = 0 ; i < m ; i++, ij++)
        {
            a[ij] = 0. ;
            for (k = 0, kj = kj0 ; k < n ; k++) a[ij] += b[ki++] * c[kj++] ;
        }
    }
    return a;
}
/*---------------------------------------------------------------------
        Produit de matrice : Amp = Bmn x [C']np
        ---------------------------------------

        Entrees:
                double *A	: Adresse de A
                double *B	: Adresse de B
                double *C	: Adresse de C
                int    m         : Nombre de lignes de A et de lignes de B
                int    n         : Nombre de colonnes de B et de colonnes de C
                int    p         : Nombre de colonnes de A et de lignes de C

    Sortie :
                A est calculée
                Retour	: Adresse de A			:

    Remarque : première et dernière dimension : celles du résultat
        A ne doit pas occuper la même zone mémoire que B ou C
----------------------------------------------------------------------*/
static double *mattmat (double *a, double *b, double *c, int m, int n, int p)
{
    int i, j, k, ij, ik, jk ;

    ij = -1 ;
    for (j = 0 ; j < p ; j++)
        for (i = 0 ; i < m ; i++)
        {
        a[++ij] = 0. ;
        for (k = 0, ik = i, jk = j ; k < n ; k++, ik += m, jk += p)
            a[ij] += b[ik]*c[jk] ;
    }
    return a;
}

/***********************************************************************
c ROUTINE CHOLL :  RESOLUTION SYSTEME LINEAIRE SYMETRIQUE PAR CHOLESKI
c
c                       index = choll(a,b,n,q)
c
c                       -1
c              [C] = [B]  *  [A]  avec [B] matrice symetrique
c
c    Remarque: Seule la partie superieure de B est utilisee.
c
c    Le resultat [C] est renvoye dans A
c
c    Entree :  n        Ordre du systeme
c              q        Nombre de seconds membres
c              a(n,q)   Valeurs des seconds membres
c              b(n,n)   Matrice symetrique du systeme
c
c    Sortie:   a(n,q)   Solution du systeme.
c              index    indicateur de validite du resultat.
c                                0 resultat ok.
c                               -1 pas de solution (diag=0)
c                                >0 : nb d'inversions de signe.
c
c    Methode : Algorithme de Cholevski lineaire sans racine carree:
c
c         L' equation  A = B . C   avec B symetrique peut s'ecrire :
c
c                      A = R'.D.R. C  avec:
c
c             R matrice triangulaire superieure a diagonale unite
c             D matrice diagonale.
c
c         La resolution du probleme
c                      A = R'.D.R.C
c         est alors ramenee a la resolution de
c                      A = R'. E
c         puis
c                      (E/D) = R . C
c
c         Les elements de la partie superieure de la matrice R sont calcules
c         et ranges dans la partie superieure de B.
c         Les elements de D sont calcules simultanement et ranges dans la
c         la diagonale de B.
c         Les elements de E sont calcules et ranges dans A.
c         Les elements de C sont calcules et ranges dans A.
c
c
*************************************************************************/
static int choll(double* a, double* b, int n, int q)
{
    int i, j, k, l, invs;
    double diag, term;
    double *pa, *pb, *p1, *p2, *p3;

    invs = 0;
    /*
Calcul des elements Rij et Dii en lieu et place de B
*/
    pb = b;                         /* b(i,i) pour i=1,... */
    for (i = 0; i < n; ++i)
    {
        diag = *pb;                  /* diag = b(i,i) */
        if (i)
        {
            p1 = b;                   /* b(k,k) pour k=1,... */
            p2 = pb -i;               /* b(k,i) pour k=1,... */
            for (k = 0; k < i; ++k)
            { /* diag = diag - b(k,k)*b(k,i)*b(k,i) */
                diag -= *p1 * *p2 * *p2;
                p1 += n + 1;
                ++p2;
            }
        }
        if (diag == 0.) return(-1);
        if (diag < 0.) ++invs;
        *pb = diag;

        p3 = pb + n;                 /* b(i,j) pour j=i+1,... */
        for (j = i+1; j < n; ++j)
        {
            term = *p3;               /* term = b(i,j) */
            if (i)
            {
                p1 = b;                /* b(k,k) pour k=1,... */
                p2 = pb - i;           /* b(k,i) pour k=1,... */
                p3 -= i;                       /* b(k,j) pour k=1,... */
                for (k = 0; k < i; ++k)
                {
                    /* term = term - b(k,i)*b(k,k)*b(k,j) */
                    term -= *p2++ * *p1 * *p3++;
                    p1 += n + 1;
                }
            }
            *p3 = term/diag;          /* b(i,j) = term / diag */
            p3 += n;
        }
        pb += n + 1;                 /* b(i+1,i+1) */
    }
    /*
calcul des elements de E en lieu et place de A
*/
    pa = a;                         /* a(1,l) pour l=1,... */
    for (l = 0; l < q; ++l)
    {
        p1 = pa;                     /* a(i,l) pour i=1,... */
        pb = b;                      /* b(1,i) pour i=1,... */
        for (i = 0; i < n; ++i)
        {
            term = *p1;               /* term = a(i,l) */
            if(i)
            {
                p2 = pb;                /* b(k,i) pour k=1,... */
                p3 = pa;                /* a(k,l) pour k=1,... */
                for (k = 0; k < i; ++k)
                    /* term = term - b(k,i)*a(k,l) */
                    term -= *p2++ * *p3++;
            }
            *p1++ = term;             /* a(i,l) = term, puis a(i+1,l)*/
            pb += n;                  /* b(1,i+1) */
        }
        /*
calcul des elements de C en lieu et place de A
*/
        p1 = pa + n - 1;             /* a(i,l) pour i=n,n-1,... */
        pb = p2;                     /* b(i,i) pour i=n,n-1,... */
        for (i = n; i > 0; --i)
        {
            term = *p1 / *pb;         /* term = a(i,l)/b(i,i) */
            if(i < n)
            {
                p2 = pb + n;            /* b(i,i+1) */
                p3 = p1 + 1;            /* a(i+1,l) */
                for (k = i+1; k <= n; k++)
                {/* term = term - b(i,k)*a(k,l) */
                    term -= *p2 * *p3++;
                    p2 += n;             /* b(i,k+1) */
                }
            }
            *p1-- = term;             /* a(i,l) = term, puis a(i-1,l)*/
            pb -= n + 1;              /* b(i-1,i-1) */
        }
        pa += n;                     /* a(1,l+1) */
    }
    return(invs);
}
#if 0
/***********************************************************************
  ROUTINE debcholl :  Préparation de la résolution par cholesky
c
c                       index = choll(b,n)
c
c    Remarque: Seule la partie superieure de B est utilisee.
c
c    Entree :  n        Ordre du systeme
c              b(n,n)   Matrice symetrique du systeme
c
c    Sortie:   b(n,n)   Solution du systeme.
c              index    indicateur de validite du resultat.
c                                0 resultat ok.
c                               -1 pas de solution (diag=0)
c                                >0 : nb d'inversions de signe.
c
c    Methode : Algorithme de Cholevski lineaire sans racine carree:
c
c         B symetrique est décomposée en R'.D.R :
c
c             R matrice triangulaire superieure a diagonale unite
c             D matrice diagonale.
c
c         Les elements de la partie superieure de la matrice R sont calcules
c         et ranges dans la partie superieure de B.
c         Les elements de D sont calcules simultanement et ranges dans la
c         la diagonale de B.
c
c
*************************************************************************/
static int debcholl(double* b, int n)
{
    int i, j, k, invs;
    double diag, term;
    double *pb, *p1, *p2, *p3;

    invs = 0;
    /*
Calcul des elements Rij et Dii en lieu et place de B
*/
    pb = b;                         /* b(i,i) pour i=1,... */
    for (i = 0; i < n; ++i)
    {
        diag = *pb;                  /* diag = b(i,i) */
        if (i)
        {
            p1 = b;                   /* b(k,k) pour k=1,... */
            p2 = pb -i;               /* b(k,i) pour k=1,... */
            for (k = 0; k < i; ++k)
            { /* diag = diag - b(k,k)*b(k,i)*b(k,i) */
                diag -= *p1 * *p2 * *p2;
                p1 += n + 1;
                ++p2;
            }
        }
        if (diag == 0.) return(-1);
        if (diag < 0.) ++invs;
        *pb = diag;

        p3 = pb + n;                 /* b(i,j) pour j=i+1,... */
        for (j = i+1; j < n; ++j)
        {
            term = *p3;               /* term = b(i,j) */
            if (i)
            {
                p1 = b;                /* b(k,k) pour k=1,... */
                p2 = pb - i;           /* b(k,i) pour k=1,... */
                p3 -= i;                       /* b(k,j) pour k=1,... */
                for (k = 0; k < i; ++k)
                {
                    /* term = term - b(k,i)*b(k,k)*b(k,j) */
                    term -= *p2++ * *p1 * *p3++;
                    p1 += n + 1;
                }
            }
            *p3 = term/diag;          /* b(i,j) = term / diag */
            p3 += n;
        }
        pb += n + 1;                 /* b(i+1,i+1) */
    }

    return(invs);
}
#endif

/**********************************************************************
c ROUTINE NextCHOLL:RESOLUTION SYSTEME LINEAIRE SYMETRIQUE PAR CHOLESKI
c
c                       index = choll(a,b,n,q)
c
c   Résolution ultérieure de [B]*[C] = [A] qui suit un premier appel
c   de la routine CHOLL.
C   A cet appel on suppose que b contient les éléments de D et R calculés
c   lors de l'appel précédent.
c   Le resultat [C] est renvoye dans A. [B] est inchangé.
c
c    Entree :  n        Ordre du systeme
c              q        Nombre de seconds membres
c              a(n,q)   Valeurs des seconds membres
c              b(n,n)   éléments de D et R calculés précédemment calculés
c
c    Sortie:   a(n,q)   Solution du systeme.
c              index    indicateur de validite du resultat.
c                                0 resultat ok.
c                               -1 pas de solution (diag=0)
c                                >0 : nb d'inversions de signe.
c
c    Methode : Algorithme de Cholevski lineaire sans racine carree:
c
c         L' equation  A = B . C   avec B symetrique peut s'ecrire :
c
c                      A = R'.D.R. C  avec:
c
c             R matrice triangulaire superieure a diagonale unite
c             D matrice diagonale.
c
c         La resolution du probleme
c                      A = R'.D.R.C
c         est alors ramenee a la resolution de
c                      A = R'. E
c         puis
c                      (E/D) = R . C
c
c         Les elements de la partie superieure de B contiennent la matrice R
c         Les elements de D dans la diagonale de B.
c         Les elements de E sont calcules et ranges dans A.
c         Les elements de C sont calcules et ranges dans A.
c
c
*************************************************************************/
static void nextcholl(double* a, double* b, int n, int q)
{
    int i, k, l;
    double term;
    double *pa, *pb, *p1, *p2, *p3;

    /*
calcul des elements de E en lieu et place de A
*/
    pa = a;                         /* a(1,l) pour l=1,... */
    for (l = 0; l < q; ++l)
    {
        p1 = pa;                     /* a(i,l) pour i=1,... */
        pb = b;                      /* b(1,i) pour i=1,... */
        for (i = 0; i < n; ++i)
        {
            term = *p1;               /* term = a(i,l) */
            if(i)
            {
                p2 = pb;                /* b(k,i) pour k=1,... */
                p3 = pa;                /* a(k,l) pour k=1,... */
                for (k = 0; k < i; ++k)
                    /* term = term - b(k,i)*a(k,l) */
                    term -= *p2++ * *p3++;
            }
            *p1++ = term;             /* a(i,l) = term, puis a(i+1,l)*/
            pb += n;                  /* b(1,i+1) */
        }
        /*
calcul des elements de C en lieu et place de A
*/
        p1 = pa + n - 1;             /* a(i,l) pour i=n,n-1,... */
        pb = p2;                     /* b(i,i) pour i=n,n-1,... */
        for (i = n; i > 0; --i)
        {
            term = *p1 / *pb;         /* term = a(i,l)/b(i,i) */
            if(i < n)
            {
                p2 = pb + n;            /* b(i,i+1) */
                p3 = p1 + 1;            /* a(i+1,l) */
                for (k = i+1; k <= n; k++)
                {/* term = term - b(i,k)*a(k,l) */
                    term -= *p2 * *p3++;
                    p2 += n;             /* b(i,k+1) */
                }
            }
            *p1-- = term;             /* a(i,l) = term, puis a(i-1,l)*/
            pb -= n + 1;              /* b(i-1,i-1) */
        }
        pa += n;                     /* a(1,l+1) */
    }
}

//------------------------------------------------------------------------
//   Le constructeur qui hérite de SimScene
//-------------------------------------------------------------------------
Sacso::Sacso()
{
    // On ajoute le choix Cde Position-Orientation ou Cde Force-Couple au menu contextuel
    actionForceOrPos = new QAction(tr("Cde en Force-Couple"), this);
    connect(actionForceOrPos, SIGNAL(triggered()), this, SLOT(ForceOrPosSlot()));
    addAction(actionForceOrPos);
    fModeXouF = X_SELECT;  // en mode Cde Position-Orientation au départ

    // On remplace le texte du Help
    delete labelHelp;
    labelHelp = new QLabel(
            tr(
                    "<h5>ACTION DES TOUCHES DU CLAVIER :</h5>"
                    "<ul><li>p(minus) pour projection orthogonale</li>"
                    "<li>P(majus) pour perspective</li></ul>"
                    "<h5>Mouvements observateur :</h5>"
                    "<ul>"
                    "<li>Touches e-E     : +/- éloignement</li>"
                    "<li>Touches Deb-Fin : Droite - Gauche</li>"
                    "<li>Page Haut-Bas   : Haut - Bas</li>"
                    "<li>Fleches &gt; - &lt; : rotation azimut</li>"
                    "<li>Fleches ^ - V   : rotation site</li>"
                    "<li>ESC             : point de vue initial</li></ul>"
                    "<h5>Mouvements maquette :</h5>"
                    "<ul><li>0,o,O: Retour origine</li>"
                    "<li>x, X : +/- Translation/Force en x</li>"
                    "<li>y, Y : +/- Translation/Force en y</li>"
                    "<li>z, Z : +/- Translation/Force en z</li>"
                    "<li>r, R : +/- Rotation/Moment en roulis(x)</li>"
                    "<li>t, T : +/- Rotation/Moment en tangage(y</li>"
                    "<li>l, L : +/- Totation/Moment en lacet (z)</li>"
                    "</ul>"
                    ), 0);
    // On change le titre de la fenetre
    labelHelp->setWindowTitle("Aide Visu Sacso");

}

//------------------------------------------------------------------------
// Réception des paramètres envoyés par le gestionnaire de la simulation.
//------------------------------------------------------------------------
void Sacso::setConfig(double *data)
{
    int i;
    for (i = 0; i < 3; i++)
    {
        OA[i] = data[i];
        Carol[i] = data[i+3];
        T[i] = data[i+6];
        T[i+3] = data[i+9];
        T[i+6] = data[i+12];
    }
    carolNok = true;
    wGL->updateGL();
}

//-----------------------------------------------------------------------------
// Le slot du menu contextuel choix Cde Position-Orientation / Cde Force-Couple
//-----------------------------------------------------------------------------
void Sacso::ForceOrPosSlot()
{
    if (fModeXouF ==  X_SELECT)
    {
        fModeXouF = F_SELECT ;
        actionForceOrPos->setText(tr("Cde en Pos/Orient."));
    }
    else
    {
        fModeXouF = X_SELECT ;
        actionForceOrPos->setText(tr("Cde en Force/Couple"));
    }
}

//-----------------------------------------------------------------------------
//   Eléments de description de la scene
//-----------------------------------------------------------------------------
#define Epouls2 0.0025
#define Epoul 0.005
static double Rpoul = 0.0125;				/* Rayon poulie */
static int flagbas[9] = {0, 1, 1, 0, 0, 1, 1, 0, 0};	/* 1 pour poulies basses */
static double OA_ref[3*4] = {-1., 0., 0., 1.,0.,0.,0., 1.,0.,0.,0.,1.}; // XYZ + Mat3x3 identité

/* Position des points d'attache des cables sur la maquette */
static double ABi[3*9] =
{
    -0.523,  0.000, -0.001,  /* 0 : point central arrière */
     0.000,  0.000,  0.041,  /* 1 : ventral tribord */
     0.000,  0.000,  0.041,  /* 2 : ventral babord */
    -0.523,  0.008, -0.005,  /* 3 : AR tribord */
    -0.523, -0.008, -0.005,  /* 4 : AR babord */
     0.000,  0.000, -0.047,  /* 5 : dorsal tribord */
     0.000,  0.000, -0.047,  /* 6 : dorsal babord */
    -0.443,  0.080, -0.003,  /* 7 : roulis tribord */
    -0.443, -0.080, -0.003   /* 8 : roulis babord */
};

/* Points droits des poulies */
static double OHi[3*9] =
{
    -3.1115000, -0.0000000,  2.8511000, /* 0 : poteau haut ventral */
    -0.1345000,  2.0160321,  2.0160321, /* 1 : poteau bas ventral tribord */
    -0.1345000, -2.0160321,  2.0160321, /* 2 : poteau bas ventral babord */
    -3.1115000,  2.7539511,  0.7379190, /* 3 : poteau haut AR tribord */
    -3.1115000, -2.7539511,  0.7379190, /* 4 : poteau haut AR babord */
    -0.1345000,  2.0160321, -2.0160321, /* 5 : poteau bas dorsal tribord */
    -0.1345000, -2.0160321, -2.0160321, /* 6 : poteau bas dorsal babord */
    -3.1115000,  0.0210925, -2.8510220, /* 7 : poteau haut roulis tribord */
    -3.1115000, -0.0210925, -2.8510220  /* 8 : poteau haut roulis babord */
};

#define DEFSITINI 34567	  /* Valeur du FlagExact pour modifier la référence  */
/*---------------------------------------------------------------------------*/
/*            Modele geometrique exact et simplifie                          */
/*---------------------------------------------------------------------------*/
/*  Entrees :                                                                */
/*      OA[3*4]  : Position-Orientation de la maquette                       */
/*                                                                           */
/*      FlagExact : 0 pour modele simplifie                                  */
/*                   (si pas encore initialisé, OA est pris pour référence)  */
/*                  1 pour modele exact                                      */
/*  Sorties :                                                                */
/*      Qi[9]     : Longueurs calculees                                      */
/*      A[6*9]    : matrice statique =-P' dans Ro                            */
/*---------------------------------------------------------------------------*/
void Sacso::Geodir(double *Qi, double *A, double *OA, int FlagExact)
{
/* flag à 1 si les constantes du modèle simplifié sont pas encore calculées  */
    double ABi_o[3], OBi[3], HiBi[3], FiBi[3], AiBi[3];
    double h, f2, f, e, l2, l, rho2, rho;
    double rctet, rstet, tetacor, Q0;
    int i, i3, i6;
    double *Com = OA+3;

    if(FlagExact !=0) /* Modele geometrique exact */
    {
        /* Io = [1;0;0];                */
        for (i=0; i< 9; i++)
        {
            i3 = 3*i; i6 = 2*i3;
            m3v3(ABi_o, Com, ABi+i3);		/* ABi_o = Com*ABi(:,i); 		*/
            addv3(OBi, OA, ABi_o);			/* OBi = OA + ABi_o; 			*/
            subv3(HiBi, OBi, OHi+i3);		/* HiBi = OBi - OHi(:,i); 		*/
            h = *HiBi;						/* h = HiBi'*Io; 				*/
            cpyv3(FiBi, HiBi); *FiBi -= h;	/* FiBi = HiBi-h*Io; 			*/
            f2 = carrev3(FiBi); f = sqrt(f2);/* f2 = FiBi'*FiBi ; 			*/
            /* scav3(Ji,1./f,FiBi);	*/		/* Ji = FiBi/f; 				*/
            e = f - Rpoul;
            l2 = e*e + h*h; l = sqrt(l2);
            rho2 = l2 - Rpoul*Rpoul; rho = sqrt(rho2);
            if (flagbas[i])
            {
                rctet = Rpoul*(Rpoul*h-rho*e)/l2;
                rstet = Rpoul*(Rpoul*e+rho*h)/l2;
                tetacor = atan2(-rctet,rstet);
                Q0 = Rpoul*(M_PI - tetacor);
            }
            else
            {
                rctet = Rpoul*(Rpoul*h+rho*e)/l2;
                rstet = Rpoul*(Rpoul*e-rho*h)/l2;
                tetacor = atan2(-rctet,rstet);
                Q0 = Rpoul*(M_PI + tetacor);
            }
            /* AiBi = HiBi - (Rpoul+rstet)*Ji - rctet*Io; */
            cumv3(AiBi, HiBi, -(Rpoul+rstet)/f, FiBi); *AiBi -= rctet;
            scav3(A+i6, -1./rho, AiBi);		/* Ui = AiBi/rho;				*/
            crossv3(A+i6+3, ABi_o, A+i6);/* matA(:,i)=-[Ui;cross(ABi_o,Ui)];*/
            Qi[i] = rho + Q0;
            if (FlagExact == DEFSITINI || 1==1) /* Ini. const. modele simplifie */
            {
                Qzero[i] = Q0;
                subv3(OAi+i3, OBi, AiBi);
                FlagNotiniConstSmp = 0; /* Ini. const. modele simplifie faite */
            }
        }
    }
    else /* Modele simplifie */
    {
        if(FlagNotiniConstSmp)
            Geodir(Qi, A, OA_ref, DEFSITINI);
        for (i=0; i< 9; i++)
        {
            i3 = 3*i; i6 = 2*i3;
            m3v3(ABi_o, Com, ABi+i3);		/* ABi_o = Com*ABi(:,i); 		*/
            addv3(OBi, OA, ABi_o);		/* OBi = OA + ABi_o; 			*/
            subv3(AiBi, OBi, OAi+i3);		/* AiBi = OBi - OAi(:,i); 		*/
            rho = longv3(AiBi);
            scav3(A+i6, -1./rho, AiBi);		/* Ui = AiBi/rho;				*/
            crossv3(A+i6+3, ABi_o, A+i6);	/* matA(:,i)=-[Ui;cross(ABi_o,Ui)];*/
            Qi[i] = rho + Qzero[i];
        }
    }
}

/*----------------------------------------------------------------------------------*/
/* glu_vertex() : Association point - texture pour les mosaïques                    */
/*----------------------------------------------------------------------------------*/
static void APIENTRY glu_vertex(GLdouble *xyz)
{
    glTexCoord2dv(xyz + 3); /* Coordonnées texture */
    glVertex3dv(xyz);		/* Coordonées points */
}

//--------------------------------------
// CONSTRUCTION DES ELEMENTS DE LA SCENE
//--------------------------------------
#define NbptAiles 13
#define NbptDerive 6
#define NbptAileron 4
void Sacso::create()
{
    double bid1[9], bid2[6*9];
    int i, i3;

    // Nombre de paramètres à transmettre par setConfig
    nbpars = 15;
    /* Initialisation diverses */
    OA[0]=-1.0;
    OA[1]=0.0;
    OA[2]=0.0;
    Carol[0]=0.0;
    Carol[1] =0.0;
    Carol[2]=0.0;
    carol2mrot(OA+3,Carol);  carolNok = false;
    Geodir(bid1, bid2, OA, 0) ; /* pour les OAi */
    FlagNotiniConstSmp = 1;
    FlagExact = 1;
    meserrlec = mesdatnok = 0;
    for(i=0; i < 6 ; i++) Fxdes[i] = 0.;

    /* couleurs */
    double colfuselage[]={70.0/255.0,130.0/255.0,180.0/255.0};
    double colgrisnoir[]= {0.15,0.15,0.015};
    double colailes[]= {70.0/255.0,130.0/255.0,180.0/255.0};
    double colaileron[]= {70.0/255.0,130.0/255.0,180.0/255.0};

    /* éclairage*/
    static GLUquadricObj *quadric;
    GLUtriangulatorObj *tess;

    /* definition/declaration des points dans repere avion (x vers av et z vers bas) */
    /* s,t : coordonnées du point correspondant de la texture compris entre (0,0) et (1,1) */
    static GLdouble Ailes[NbptAiles][5] =  /* Points des Ailes */
    {
        /* { x, y, z, s, t } */
        {   0.00,   0.00, 0.0,  0.00,  0.40 },
        {  0.207, -0.073, 0.0, -0.25,  0.00 },
        { -0.213, -0.333, 0.0, -1.00,  0.80 },
        { -0.213, -0.063, 0.0, -0.15,  0.80 },
        { -0.320, -0.063, 0.0, -0.15,  0.95 },
        { -0.340, -0.044, 0.0, -0.10,  1.00 },
        { -0.213, -0.043, 0.0, -0.10,  0.87 },
        { -0.213,  0.043, 0.0,  0.10,  0.87 },
        { -0.340,  0.044, 0.0,  0.10,  1.00 },
        { -0.320,  0.063, 0.0,  0.15,  0.95 },
        { -0.213,  0.063, 0.0,  0.15,  0.80 },
        { -0.213,  0.333, 0.0,  1.00,  0.80 },
        {  0.207,  0.073, 0.0,  0.25,  0.00 },
    };
    static GLdouble Derive[NbptDerive][5] =  /* Points de la dérive */
    {
        /* { x, y, z, s, t } */
        { -0.063, 0.0, -0.042, 1.5, 0.0 },
        { -0.137, 0.0, -0.064, 1.2, 0.1 },
        { -0.36 , 0.0, -0.230, 0.4, 1.0 },
        { -0.417, 0.0, -0.230, 0.05, 1.0},
        { -0.377, 0.0, -0.064, 0.2, 0.1 },
        { -0.377, 0.0, -0.044, 0.2, 0.0 }
    };
    static GLdouble    Aileron_G[NbptAileron][5] =  /* Points ailerons gauche */
    {
        /* { x, y, z, s, t } */
        { -0.213,  -0.063, 0.0,  0.0,  0.0 },
        { -0.213,  -0.333, 0.0,  1.0,  0.5 },
        { -0.267,  -0.340, 0.0,  1.0, 1.0 },
        { -0.287,  -0.063, 0.0,  0.0, 1.0 },
    };
    static GLdouble    Aileron_D[NbptAileron][5] = /* Points ailerons droit */
    {
        /* { x, y, z, s, t } */
        { -0.287,  0.063, 0.0,  0.0, 1.0 },
        { -0.267,  0.340, 0.0,  1.0, 1.0 },
        { -0.213,  0.333, 0.0,  1.0,  0.5 },
        { -0.213,  0.063, 0.0,  0.0,  0.0 },
    };
    double x,y,z,teta, ct, st, dx, dy;

    /* Création objets*/
    quadric = gluNewQuadric();/* Nouvel objet quadric générique */
    gluQuadricTexture(quadric, GL_FALSE);/* qui ne sera pas éventuellemnt texturé */

    tess = gluNewTess();/* Nouvel objet mosaïque générique */
#define _GLUfuncptr void (CALLBACK *)()
    gluTessCallback(tess, GLU_TESS_BEGIN, (_GLUfuncptr) glBegin);/* Begin standard */
    gluTessCallback(tess, GLU_TESS_END, glEnd);/* End standard */
    /* Vertex modifié */
    gluTessCallback(tess, GLU_TESS_VERTEX, (_GLUfuncptr) glu_vertex);

    /* Construction du corps du Mirage par une DisplayList */
    DL_Corps = glGenLists(1);/* Un numéro de DisplayList */
    glNewList(DL_Corps, GL_COMPILE);/* Définition sans execution */
    /* Positionnement dans repere OpenGL */
    glPushMatrix(); /* Push repere Avion */
    glRotatef(-90., 0.0, 0.1, 0.0);
    /* On passe en repère Z arrière pour les cylindres */
    /* La verrière : ovale allongé en z  */
    glPushMatrix();
    glColor3d(0.5, 0.5, 1.0);
    glTranslated(-0.018, 0.0, -0.340);
    glScaled(1.0, 1.0, 3.);
    gluSphere(quadric, 0.03, 10, 20);
    glPopMatrix();
    /* Cone du nez */
    glColor3dv(colfuselage);
    glPushMatrix();
    glTranslated(0.0, 0.0, -0.62);
    gluCylinder(quadric, 0.0, 0.033, 0.177, 20, 2);
    glPopMatrix();
    /* Entrées d'air */
    glColor3dv(colfuselage);
    glPushMatrix();
    glTranslated(0.0, 0.0, -0.247);
    glScaled(1.0, 2.2, 1.0);
    gluCylinder(quadric, 0.033, 0.047/2.2, 0.447, 20, 2);
    glPopMatrix();
    /* Cones entrée d'air */
    glColor3dv(colgrisnoir);
    glTranslated(0.0, 0.0, -0.483);
    glScaled(1.0, 2.2, 1.0);
    gluCylinder(quadric, 0., 0.057/2.2, 0.267, 20, 2);
    glPopMatrix(); /* Retour repère avion */
    /* Aile principale */
    glColor3dv(colailes);
    glNormal3d(0.0, 0.0, -1.0);
    gluBeginPolygon(tess);/* Début polygone mosaïque aile */
    for (i = 0; i < NbptAiles; i ++)
        gluTessVertex(tess, Ailes[i], Ailes[i]);  /* Les points */
    gluEndPolygon(tess);/* Fin polygone mosaïque aile */
    /* La dérive */
    glColor3dv(colailes);
    glNormal3d(0.0, 1.0, 0.0);
    gluBeginPolygon(tess);/* Début polygone mosaïque dérive */
    for (i = 0; i < NbptDerive; i ++)
        gluTessVertex(tess, Derive[i], Derive[i]);/* Les points */
    gluEndPolygon(tess);/* Fin polygone mosaïque dérive */
    /* Extérieur  */
    glPushMatrix();
    glRotatef(-90., 0.0, 0.1, 0.0); /* On repasse en repère Z arrière  */
    /* la poutre */
    glTranslated(0.0, 0.0, 0.340);
    glColor3d(0.8, 0.5, 0.0);
    gluCylinder(quadric, 0.022, 0.008, -ABi[0]-0.340-0.005, 4, 2);
    /* Le réacteur (non texturé) */
    glColor3dv(colgrisnoir);
    gluCylinder(quadric, 0.047, 0.040, 0.08, 20, 2);
    glPopMatrix(); /* Repere Avion */
    /* Les T */
    glPushMatrix();
    /* grand T */
    glColor3d(0.8, 0.5, 0.0);
    glTranslatef(ABi[7*3+0], 0.,0. );
    glRotatef(90., 1., 0., 0.);
    gluCylinder(quadric, 0.01, 0.01, ABi[3*7+1], 8, 2);
    glRotatef(180., 1., 0., 0.);
    gluCylinder(quadric, 0.01, 0.01, ABi[3*7+1], 8, 2);
    glPopMatrix(); /* Repere Avion */
    /* petit T */
    glPushMatrix();
    glTranslatef(ABi[3*3+0], 0.,0. );
    glRotatef(90., 1., 0., 0.);
    gluCylinder(quadric, 0.01, 0.01, ABi[3*3+1], 8, 2);
    glRotatef(180., 1., 0., 0.);
    gluCylinder(quadric, 0.01, 0.01, ABi[3*3+1], 8, 2);
    glPopMatrix();
    /* Cylindre fuselage */
    glPushMatrix();
    glRotatef(-90., 0.0, 0.1, 0.0); /* On repasse en repère Z arrière  */
    glColor3dv(colfuselage);
    glTranslated(0.0, 0.0, -0.443);
    gluCylinder(quadric, 0.04-0.007 , 0.04+0.007, 0.783, 20, 2);
    glPopMatrix(); /* Retour repere Avion */
    glEndList();/* fin de  DL_Corps */

    /* Construction Aileron gauche par une DisplayList */
    DL_Aileron_G = glGenLists(1);
    glNewList(DL_Aileron_G, GL_COMPILE);
    {
        glColor3dv(colaileron);
        glNormal3d(0.0, 1.0, 0.0);
        gluBeginPolygon(tess);
        for (i = 0; i < NbptAileron; i ++)
            gluTessVertex(tess, Aileron_G[i], Aileron_G[i]);
        gluEndPolygon(tess);
    }
    glEndList();/* fin DL_AILERON_G */

    /* Construction Aileron droit par une DisplayList */
    DL_Aileron_D = glGenLists(1);
    glNewList(DL_Aileron_D, GL_COMPILE);
    {
        glColor3dv(colaileron);
        glNormal3d(0.0, 1.0, 0.0);
        gluBeginPolygon(tess);
        for (i = 0; i < NbptAileron; i ++)
            gluTessVertex(tess, Aileron_D[i], Aileron_D[i]);
        gluEndPolygon(tess);
    }
    glEndList();/* fin DL_AILERON_D */

    /* Construction cone pour flèches dans DisplayList */
    DL_Cone = glGenLists(1);
    glNewList(DL_Cone, GL_COMPILE);
    gluCylinder(quadric, 0.02, 0.0, 0.05, 4, 2);
    glEndList();/* fin DL_Cone */

    /* Construction Veine (Rv veine avec z est vers le haut) */
    DL_Veine = glGenLists(1);
    glNewList(DL_Veine, GL_COMPILE);
    /* Le cylindre intérieur de 4m de diamètre */
    glColor3d(0.3,0.4,0.5);				/* foncé */
    glTranslated(0., 0., -1.0);
    gluCylinder(quadric, 2., 2., 1.0, 40, 2);
    glTranslated(0., 0., 1.0) ;			/* Retour Rv */
    /* Le marche-pied 4m à 5m de diamètre */
    glColor3d(0.5,0.9,0.5);				/* clair */
    gluDisk(quadric, 2.0, 2.5,40,2);
    /* Le cylindre de la marche (le haut est au niveau 0) */
    glTranslated(0., 0., -0.1);
    glColor3d(0.3,0.4,0.5); /* foncé */
    gluCylinder(quadric, 2.5, 2.5, 0.1, 40, 2);
    /* Le sol extérieur (au niveau du bas de la marche à -0.1) */
    glColor3d(0.9,0.9,0.9); /* clair */
    gluDisk(quadric, 2.5, 8.,40,2);
    glTranslated(0., 0., 0.1);			/* Retour Rv */
    /* Les poteaux  */
#define Rsocle (0.35/M_SQRT2)	/* rayon socle */
#define Cs2pot 0.08				/* demi coté poteau */
#define Rpot   (Cs2pot*M_SQRT2)	/* rayon poteau */
#define AxeHi  (2.*Cs2pot + 0.02) /* distance axe tangentiel poulie - centre poteau */
#define Hpotbas 1.38
#define Rmot 0.035
#define Raxep 0.001             /* rayon axe poulie */
#define Laxep 0.1               /* longueur axe poulie */
    for(i=i3 = 0; i < 8; i++, i3 += 3)
    {
        glPushMatrix();
        /* Les axes verticaux des poulies */
        x = OHi[i3+2]; y = OHi[i3+1]; z = -OHi[i3];
        glTranslated(x, y, z-Laxep/2.);
        glColor3d(0.1,0.1,0.1);
        gluCylinder(quadric, Raxep, Raxep, Laxep, 4, 2);
        glPopMatrix();
        if (i == 7)
        {
            /* Le dernier axe */
            glPushMatrix();
            /* Les axes verticaux des poulies */
            x = OHi[i3+5]; y = OHi[i3+4]; z = -OHi[i3+3];
            glTranslated(x, y, z-Laxep/2.);
            glColor3d(0.1,0.1,0.1);
            gluCylinder(quadric, Raxep, Raxep, Laxep, 4, 2);
            glPopMatrix();
            x = 0.5*(OHi[i3+2]+OHi[i3+5]); y = 0.5*(OHi[i3+1]+OHi[i3+4]);
        }

        glPushMatrix();
        /* Le socle */
        teta = atan2(y,x);
        ct = cos(teta); st = sin(teta);
        dx = AxeHi*ct; dy = AxeHi*st;
        glTranslated(x + dx , y + dy, -0.07);	/* Niveau -0.07 */
        glRotatef(teta*RAD2DEG+45., 0., 0., 1.);
        /* la marche */
        glColor3d(0.2,0.4,0.3);
        gluCylinder(quadric, Rsocle, Rsocle, 0.07, 4, 2);
        glTranslated(0. , 0., 0.07);			/* Niveau 0 */
        /* Le dessus du socle */
        gluDisk(quadric, 0., Rsocle, 4, 2);
        /* Le poteau */
        if (i == 0) glColor3d(0.7,0.9,0.8);
        else glColor3d(0.9,0.8,0.7);
        z = flagbas[i] ? Hpotbas : -OHi[i3]+0.15;
        gluCylinder(quadric, Rpot, Rpot, z, 4, 2);
        /* Le dessus du poteau */
        glTranslated(0. , 0., z);
        gluDisk(quadric, 0.0, Rpot, 4,2);
        /* Le moteur */
        z = -z + (flagbas[i] ? Hpotbas-0.15 : 2.14);
        glRotatef(-45., 0., 0., 1.);
        glTranslated(-(Cs2pot+Rmot) , 0., z);
        glRotatef(90., 1., 0., 0.);
        if (i < 7) gluCylinder(quadric, Rmot, Rmot, 0.35, 12, 2);
        else
        {
            gluCylinder(quadric, Rmot, Rmot, 0.35, 12, 2);
            glRotatef(180., 1., 0., 0.);
            gluCylinder(quadric, Rmot, Rmot, 0.35, 12, 2);
        }
        glPopMatrix();
    }
    glEndList();/* fin DL_Veine */

    /* Construction Poulie */
    DL_Poulie = glGenLists(1);
    glNewList(DL_Poulie, GL_COMPILE);
    glColor3d(1.,0.2,0.2);
    glTranslated(0. , 0., -Epouls2);
    gluCylinder(quadric, Rpoul, Rpoul, Epoul, 20, 2);
    glTranslated(0. , 0., -Laxep/2.);
    glColor3d(0.1,0.1,0.1);
    gluCylinder(quadric, Raxep, Raxep, Laxep, 4, 2);
    glEndList();/* fin DL_Poulie */

    gluDeleteTess(tess);
    gluDeleteQuadric(quadric);

    /* Couleur du fond */
    glClearColor(0.85, 0.95, 0.95, 1.0);

}

#if 0
/*----------------------------------------------------------------------------------*/
/* Utilitaire conversion situation MLL -> Situation OpenGL                          */
/*----------------------------------------------------------------------------------*/
static double *sit2mat44(double *m44, double *sit)
{
    m44[0] = sit[3] ; m44[4] = sit[6] ;  m44[8]  = sit[ 9] ; m44[12] = sit[0] ;
    m44[1] = sit[4] ; m44[5] = sit[7] ;  m44[9]  = sit[10] ; m44[13] = sit[1] ;
    m44[2] = sit[5] ; m44[6] = sit[8] ;  m44[10] = sit[11] ; m44[14] = sit[2] ;
    m44[3] = 0.     ; m44[7] = 0.     ;  m44[11] = 0.      ; m44[15] = 1. ;
    return m44 ;
}
#endif

/*-----------------------------------------------------------------
 Distance minimale entre deux segments de droite AB et CD
 définis par les points A, B, C, D
 Utilisé pour vérifier la non collision des cables
--------------------------------------------------------------------*/
double dmin2segs(double *A, double *B, double *C, double *D)
{
    double BA[3], CD[3], AC[3];
    double x,y,a,b,c,d,e,f, delta;
    double aa, bb, cc;

    subv3(BA,A,B);
    subv3(CD,D,C);
    subv3(AC,C,A);
    /*
% On commence par résoudre
%    a x + c y + d = 0
%    c x + b y + e = 0
% Avec :
*/
    a = dotv3(BA,BA);
    b = dotv3(CD,CD);
    c = dotv3(BA,CD);
    d = dotv3(BA,AC);
    e = dotv3(AC,CD);
    f = dotv3(AC,AC);
    delta = a*b-c*c;
    if (delta != 0.)
    {
        x = (c*e-b*d)/delta;
        if (x >= 0. && x <= 1.)
        {
            y = (c*d-a*e)/delta;
            if (y >= 0. && y <= 1.) return sqrt(f + d*x + e*y);
            if (y > 1.)
            { y = 1. ; aa = a; bb = c+d ; cc = f+b+e+e; }/* ???? Pas sur ???? */
            else {y = 0. ; aa = a; bb = d ;   cc = f; }  /* ???? Pas sur ???? */
            if (aa != 0.) x = -bb/aa ; else x = 0.;
            if (x >= 0.)
            {
                if (x <= 1.)return sqrt(cc+bb*x);
                x = 1.; aa = b; bb = c+e; cc = f + a + d+d;
                if (aa != 0.) y = -bb/aa ; else y = 0.;
                if (y >= 0.)
                {/* Cas impossible */
                    if (y <= 1.) return sqrt(cc+bb*y);
                    return sqrt(a+b+f+2.*(c+d+e));
                }
                return sqrt(a+d+d+f);
            }
            x = 0.; aa = b; bb = e; cc = f;
            if (aa != 0) y = -bb/aa ; else y = 0.;
            if (y >= 0.)
            {/* Cas impossible */
                if (y <= 1.) return sqrt(cc+bb*y);
                return sqrt(b+e+e+f);
            }
            return sqrt(f);
        }
        else
        {
            if (x > 1)
            {x = 1. ; aa = b; bb = c+e; cc = f + a + d+d; } /* ???? Pas sur ???? */
            else {x = 0. ; aa = b; bb = e;   cc = f; } /* ???? Pas sur ???? */
            if (aa != 0) y = -bb/aa ; else y = 0.;
            if (y >= 0.)
            {
                if (y <= 1.) return sqrt(cc+bb*y);
                y = 1. ; aa = a; bb = c+d ; cc = f+b+e+e;
                if (aa != 0.) x = -bb/aa ; else x = 0.;
                if (x >= 0.)
                {
                    if (x <= 1.) return sqrt(cc+bb*x);
                    return sqrt(a+b+f+2.*(c+d+e));
                }
                return sqrt(b+e+e+f);
            }
            y = 0. ; aa = a; bb = d ; cc = f;
            if (aa != 0.) x = -bb/aa ; else x = 0.;
            if (x >= 0.)
            {
                if (x <= 1.) return sqrt(cc+bb*x);
                return sqrt(a+d+d+f);
            }
            return sqrt(f);
        }
    }
    else /* delta < 0 */
    {
        double AB2, ABAC, AC2, ABAD, bid[3];
        x = 0 ; y = 0;
        AB2 = a; ABAC = -d; AC2 = f;
        /* Test si C entre A et B */
        ABAD = dotv3(BA,subv3(bid,D,A));
        if ((ABAC >= 0. && ABAC <= AB2) || (ABAD >= 0 && ABAD <= AB2))
            return sqrt(AC2-ABAC*ABAC/AB2);
        /* Test si C et D de part et d'autre de A et B */
        if ((ABAC < 0. && ABAD > AB2) || (ABAD < 0 && ABAC > AB2))
            return sqrt(AC2-ABAC*ABAC/AB2);
        if (ABAC < 0.)
        {
            if (ABAD < ABAC) return sqrt(AC2);
            else return sqrt(b+e+e+f); /* AD2 (x=0, y=1) */
        }
        else
        {
            if (ABAD > ABAC) return sqrt(a+d+d+f); /* BC2 (x=1, y=0) */
            else return sqrt(a+b+f+2.*(c+d+e)) ; /* BD2 (x=1, y=1) */
        }
    }
}

#define NVAR  9	/* Programmation specifique 9 variables */
#define NEQ 6	/* Soumises à 6 contraintes égalités    */
#define NFREE 3 /* Il y a donc 3 degrés de liberté      */
#define MAXCAS 987

int Statinv (double *Fq, double *Fxdes, double *A, double *tmin, double *tmax,
             double *tref, double *tpon, int Satur)
/*---------------------------------------------------------------------------*/
/*       Modele statique inverse : Résolution de Fxdes = A * Fq              */
/*---------------------------------------------------------------------------*/
/*               (Redondant -> Prog Quadratique)                             */
/*	Sorties :							     */
/*		Fq		: Vecteur des tensions des 9 câbles	     */
/*									     */
/*	Entrées :							     */
/*		Fxdes : Torseur cartésien à réaliser			     */
/*		A     : Matrice 6 x 9  telle que  Fxdes = A * Fq             */
/*									     */
/*		tmin  : Vecteur des 9 tensions minimales		     */
/*		tmax  : Vecteur des 9 tensions maximales		     */
/*		tref  : Vecteur des 9 tensions de référence		     */
/*		tpon  : Vecteur des 9 tensions de pondération		     */
/*                                                                           */
/*              Satur : 0 : tmin et tmax ne sont pas pris en compte dans     */
/*			le calcul des Fq. Si une ou plusieurs tensions       */
/*			sont hors domaine, elle est saturée. Fxdes ne	     */
/*			sera pas réalisé, même s'il existe une solution      */
/*			qui réalise Fxdes en respectant les contraintes.     */
/*			1 : tmin et tmax sont pris en compte dans le calcul  */
/*			des Fq. Si une solution qui réalise Fxdes en	     */
/*			respectant les contraintes existe, elle est trouvée. */
/*									     */
/*    La solution minimise la somme des [Fq(i)-tref(i))/tpon(i)]^2           */
/*      tout en respectant (si Satur != 0)  tmin(i) < Fq(i) < tmax(i)        */
/*                                                                           */
/*	Retour :							     */
/*				0 : solution exacte, 0 contraintes saturées  */
/*			n > 0 : solution exacte, n contraintes saturées	     */
/*			n < 0 : solution fausse, |n| contraintes saturées    */
/*			  -10 : Echec résolution. Pas de solution.	     */
/*									     */
/*---------------------------------------------------------------------------*/
{
    int i, ii, j, jj, ij, k, jns, jNvar;
    int nbcascur, nbcasnew, nsat, icasat;
    int idem, flagtrouve, flagnew;
    int flagVK[NVAR] = {0};
    int bufindex1[MAXCAS], bufindex2[MAXCAS];
    int pixcur[NFREE], *pixsat, *pixsatcur, *pixsatnew, *pixcascur, *pixcasnew;
    double VK[NVAR*NVAR], tlim[NVAR], MM[NFREE*NFREE], mu[NFREE];
    double AD1[NEQ*NVAR], dF[NVAR], Mat[NEQ*NEQ], Fqref[NVAR];
    double delta, temp ;
    if (!Satur)  /* On ne prend pas les contraintes en compte */
    {
        matmat(dF,A,tref,NEQ,NVAR,1);
        subvec(dF,Fxdes,dF,NEQ);
        for (i = 0; i<NVAR; i++) scavec(AD1+i*NEQ,tpon[i],A+i*NEQ, NEQ);
        mattmat(Mat,AD1,AD1,NEQ,NVAR,NEQ);
        if (choll(dF,Mat,NEQ,1) < 0) return -10;
        tmatmat(Fq, AD1,dF,NVAR,NEQ,1);
        mulvec(Fq,tpon,Fq,NVAR);
        addvec(Fq,tref,Fq,NVAR);
        for (nsat = 0, i = 0; i<NVAR; i++)
        {
            if (Fq[i] < tmin[i]) { nsat--; Fq[i] = tmin[i];}
            else if (Fq[i]>tmax[i]) { nsat--; Fq[i] = tmax[i];}
        }
        return nsat;
    }

    /* Solution avec prise en compte des contraintes. Algorithme perso de
       programmation quadratique adapté de la méthode de Theil et Van de Panne */

    pixsatcur = bufindex1;/* Les différents cas de cont. saturées en cours d'examen */
    pixsatnew = bufindex2;   /* Les nouveaux cas de contraintes saturées a examiner */
    nbcascur = 1;   /* Au début c'est le cas sans contrainte qui est le cas courant */
    nbcasnew = 0;  /* Nouveaux cas d'ensembles de contraintes simultanées à saturer */
    flagtrouve = 0;                                /* On n'a pas trouvé la solution */
    nsat = 0;	                     /* Nombre decontraintes simultanément saturées */

    /* On traite les ensembles de contraintes saturées nsat par nsat (0 par 0, 1 par 1, ...) */
    for (nsat = 0; nsat <= NFREE; nsat++)
    {
        /* Examen de toutes les cas de contraintes saturées */
        for (icasat = 0, pixcascur = pixsatcur, pixcasnew = pixsatnew;
             icasat < nbcascur; icasat++,   pixcascur += nsat)
        {
            /* Calcul de la solution */
            if (nsat == 0)
            { /* Solution nominale sans contraintes inégalité */;
                matmat(dF,A,tref,NEQ,NVAR,1);
                subvec(dF,Fxdes,dF,NEQ);
                for (i = 0; i<NVAR; i++) scavec(AD1+i*NEQ,tpon[i],A+i*NEQ, NEQ);
                mattmat(Mat,AD1,AD1,NEQ,NVAR,NEQ);
                if (choll(dF,Mat,NEQ,1) < 0) return -10 ;
                tmatmat(Fq, AD1,dF,NVAR,NEQ,1);
                mulvec(Fq,tpon,Fq,NVAR);
                addvec(Fq,tref,Fq,NVAR);
                cpyvec(Fqref,Fq,NVAR);
            }
            else /* Calcul de la solution générale nsat + NEQ <= NVAR */
            {
                for (jj = jns = 0; jj < nsat; jj++, jns += nsat)
                {
                    j =  pixcascur[jj];
                    if (j < NVAR) tlim[j] = tmin[j];
                    else {j -= NVAR; tlim[j] = tmax[j];}
                    jNvar = j*NVAR;
                    pixcur[jj] = j;
                    /* Calcul des Vk si pas deja calculés */
                    if (flagVK[j]==0)
                    {
                        cpyvec(dF,AD1+j*NEQ,NEQ);
                        nextcholl(dF,Mat,NEQ,1);
                        tmatmat(VK+jNvar, AD1,dF,NVAR,NEQ,1);
                        flagVK[j] = 1;
                    }
                    /* Constitution matrice facteur de mu */
                    for (ii = 0 ; ii <= jj; ii++) /* Seule la partie supérieure est nécessaire */
                    {
                        i = pixcur[ii];
                        ij = ii+jns;
                        MM[ij] = -VK[i+jNvar];
                        if (ii==jj) MM[ij] += 1.;
                    }
                    mu[jj] = (Fqref[j] - tlim[j])/tpon[j];
                }
                if (nsat==1)
                {
                    if (MM[0]==0.) break; /* On laisse tomber ce cas */
                    mu[0] /= MM[0];
                }
                else
                {
                    if (nsat == 2)
                    {
                        delta = MM[0]*MM[3]-MM[2]*MM[2];
                        if (delta == 0.) break ; /* On laisse tomber ce cas */
                        temp  = ( MM[3]*mu[0]-MM[2]*mu[1])/delta;
                        mu[1] = (-MM[2]*mu[0]+MM[0]*mu[1])/delta;
                        mu[0] = temp;
                    }
                    else
                    {
                        if (choll(mu,MM,nsat,1) < 0) break ; /* On laisse tomber ce cas */
                    }
                }
                cpyvec(Fq, Fqref, NVAR);
                for (ii = 0 ; ii < nsat; ii++)
                {
                    i = pixcur[ii];
                    mulvec(dF,tpon,VK+i*NVAR,NVAR);
                    cumvec(Fq, Fq, mu[ii],dF,NVAR);
                }
                for (ii = 0 ;ii < nsat;ii++)
                {
                    i = pixcur[ii];
                    Fq[i] = tlim[i];
                }
            }
            flagtrouve = 1;
            /* Test si solution admissible */
            for (i=0; i < NVAR ; i++)
            {
                if (Fq[i] < tmin[i]) ii = i;
                else if (Fq[i] > tmax[i]) ii = i + NVAR;
                else continue;
                flagtrouve = 0;
                if (nsat == NFREE) {nbcasnew++; break; } /* Inutile d'en faire plus. Pb trop contraint */
                /* Constitution d'un nouveau cas de contrainte à saturer */
                for (j = 0, k= 0 ; k < nsat; k++) /* rang de la nlle contrainte / au cas en cours */
                {
                    if (ii < pixcascur[k]) break; else j++;
                }
                /* On range le nouveau cas */
                for (k=0; k < j; k++) pixcasnew[k] = pixcascur[k];
                pixcasnew[j] = ii;
                for (k = j+1; k<= nsat; k++) pixcasnew[k] = pixcascur[k-1];

                /* Test si le cas est vraiment nouveau */
                flagnew = 1 ;
                for (j = 0, pixsat = pixsatnew ; j < nbcasnew; j++, pixsat += nsat+1)
                {
                    idem = 1;
                    for (k = 0 ; k <= nsat ; k++)
                    {
                        if(pixsat[k] != pixcasnew[k]) {idem = 0; break;}
                    }
                    if (idem) {flagnew = 0; break;}
                }
                /* C'est un nouveau cas */
                if (flagnew)
                {
                    nbcasnew++;
                    pixcasnew += nsat+1;
                }
            }

            /* Test si solution optimale pour nsat > 1 (pour 0 et 1 c'est forcément optimal) */
            /* Placé après la conformité, car ce test est quasiment toujours vérifié, il sera
                           donc moins souvent appelé ici, dans la mesure ou le précédent est moins
                           fréquemment vérifié */
            if (flagtrouve && nsat > 1) for(k=0; k < nsat; k++)
            {
                if (pixcascur[k] < NVAR) { if (mu[k] > 0.) { flagtrouve = 0; break; }}
                else { if (mu[k] < 0.) { flagtrouve = 0; break; }}
            }

            if (flagtrouve) break;
        } /* Fin boucle sur icasat */
        if (flagtrouve) break;
        /* On permute les buffeurs d'index */
        pixsat = pixsatcur;
        pixsatcur = pixsatnew;
        pixsatnew = pixsat;
        nbcascur = nbcasnew;
        nbcasnew = 0;
    }
    if (flagtrouve) return nsat ;

    /* Rien trouvé. On sature la solution initiale */
    cpyvec(Fq,Fqref,NVAR);
    for (nsat = 0, i = 0; i<NVAR; i++)
    {
        if (Fq[i] < tmin[i]) { nsat--; Fq[i]=tmin[i];}
        else if (Fq[i]>tmax[i]) { nsat--; Fq[i]=tmax[i];}
    }
    return nsat ;
}

/* Pour le test réalisabilté en local */
static double tmax[9] = {120.,240.,240.,120.,120.,240.,240.,120.,120.};
static double tpon[9] = {120.,240.,240.,120.,120.,240.,240.,120.,120.};
static double tref[9] = {36.,72.,72.,36.,36.,72.,72.,36.,36.};
static double tmin[9] = {10.,10.,10.,10.,10.,10.,10.,10.,10.};
/* Messages d'erreur */
static char txterrlec[] = "  ECHEC OUVERTURE FICHIER";
static char txtdatnok[] = "  FICHIER FORMAT INCORRECT";
/*----------------------------------------------------------------------------------*/
/* Callback dessinant la scene                                                      */
/*----------------------------------------------------------------------------------*/
void Sacso::redraw(void)
{
    QFont font;
    font.setPointSize(8); // Pour les textes 2D.
    double Astat[6*9], di[9];
    double mat44[16], *Com = OA+3, *ui = Astat;
    int i, i3, i6, ibas , ihaut;
    char texte[256];
    double wwlong, angleww, ww[3];
    double OTi[27], ABi_o[27], OBi_o[27];

    GLfloat pos_lumiere1[]={1.,0.,1.,0.}; /* selon bissectrice XZ */
    /**********************************/
    /* L'éclairage                    */
    /**********************************/
    mlSetLight(GL_LIGHT0, pos_lumiere1);


    /* Calul longueur câbles et directions (dans Astat) */
    if (carolNok) {carol2mrot(OA+3,scav3(ww,DEG2RAD,Carol)); carolNok = false;}
    Geodir(di, Astat, OA, FlagExact) ;

    /* Position point d'ancrages mobiles dans Ro */
    matmat(ABi_o,Com,ABi,3,3,9);	/* calcul ABi_o */
    for(i3 = 0; i3 < 27; i3 += 3)	/* calcul OBi_o */
    {
        OBi_o[i3]   = OA[0] + ABi_o[i3] ;
        OBi_o[i3+1] = OA[1] + ABi_o[i3+1] ;
        OBi_o[i3+2] = OA[2] + ABi_o[i3+2] ;
    }    

    /* La soufflerie  */
    glTranslated(0.,0., -1.); // On baisse le sol par / au centre ecran
    glCallList(DL_Veine);
    glRotatef(90., 0.0, 1.0, 0.0);  /* Repère Saxo X vers le bas */
    glPushMatrix(); /* On push le repere Sacso */
    /* Transformations pour situation dans repere avion */
//    sit2mat44(mat44,OA);
    posEtmat2GLmat44(mat44, OA, OA+3);
    glMultMatrixd(mat44); /* On est dans le repere avion */
    glCallList(DL_Corps);
    glPopMatrix(); /* Retour repere Sacso */

    /* Les cables, tensions et poulies */
    for (i=i3=0; i < 9 ;i++, i3 += 3)
    {
        /* Les cables sans fioriture */
        glColor3d(0.1,0.,0.);
        glLineWidth(1.);
        glBegin(GL_LINES);
        glVertex3dv(&OBi_o[i3]);
        glVertex3dv(&OAi[i3]);
        glEnd() ;
        /* Sur-épaisseur et flèche pour les tensions */
        /* OTi = OBi + Ti.ui_o */
        i6 = i3+i3;
        cumv3(&OTi[i3], &OBi_o[i3], T[i]*di[i]/250., &ui[i6]);
        glLineWidth(2.5);
        glColor3d(0.,0.,0.7);
        glBegin(GL_LINES);
        glVertex3dv(&OBi_o[i3]);
        glVertex3dv(&OTi[i3]);
        glEnd();
        /* Les poulies */
        glPushMatrix(); /* Push rep. Sacso */
        ww[0]= 0.0;
        ww[1]= ui[i6+1];
        ww[2]= ui[i6+2];
        wwlong = lunitv3(ww,ww);
        angleww = atan2(ww[2], ww[1])*RAD2DEG;
        glTranslated(OHi[i3],  OHi[i3+1]-Rpoul*ww[1], OHi[i3+2]-Rpoul*ww[2]);
        glRotated(angleww,1.,0.,0.);
        glCallList(DL_Poulie);
        glPopMatrix(); /* Retour repere SACSO */
    }
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    /*------------------------------------------------*/
    /*            Affichage des textes                */
    /*------------------------------------------------*/
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, 1.*wGL->width1, 0.0, 1.*wGL->height1, -1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glColor3d(0., 0.,0.);

    ihaut = 0;
    i = sprintf(texte,"    X = %+5.3lf  Y = %+5.3lf  Z = %+5.3lf",
                OA[0], OA[1], OA[2]);
    sprintf(texte+i," Cap = %+5.1lf  Assiette = %+5.1lf Roulis = %+5.1lf",
            Carol[0], Carol[1], Carol[2]);
    ihaut += 15; wGL->renderText (0, ihaut, texte, font) ;


    ibas = wGL->height1-3; strcpy(texte, "LKK");for (i=0;i<3;i++) texte[i] += 1;
    wGL->renderText (wGL->width1 - 24, ibas, texte, font) ;
    ibas -= 15;

    if(meserrlec)
    {
        meserrlec = 0; ibas -= 15; wGL->renderText (0, ibas, txterrlec, font) ;
    }
    if(mesdatnok)
    {
        mesdatnok = 0; ibas -= 15; wGL->renderText (0, ibas, txtdatnok, font) ;
    }

    if (devInput == LOCAL)
    {
        /* Contact entre cables avant et fuselage si Uz_m change de signe */
        for (i = 1; i <= 2; i++) if (dotv3(ui+6*i, Com+6) <= 0.)
        {
            sprintf(texte,"    ATTENTION CONTACT CABLE %d ET FUSELAGE",i+1);
            ibas -= 15 ; wGL->renderText (0, ibas, texte, font) ;
        }
        for (i = 5; i <= 6; i++) if (dotv3(ui+6*i, Com+6) >= 0.)
        {
            sprintf(texte,"    ATTENTION CONTACT CABLE %d ET FUSELAGE",i+1);
            ibas -= 15 ; wGL->renderText (0, ibas, texte, font) ;
        }

        /* Contact entre câbles 1-4 ou 1-5  si leur distance devient inférieure
                       au demi entre axe en y des attaches 4-5 */
        for (i = 3; i <= 4; i++)
        {
            if (ABi[3*3+1] > dmin2segs(OAi, OBi_o, OAi+i*3, OBi_o+i*3))
            {

                sprintf(texte,"    ATTENTION CONTACT ENTRE CABLES 1 ET %d",i+1);
                ibas -= 15 ; wGL->renderText (0, ibas, texte, font) ;
            }
        }

        /* Contact entre câbles 4-8 ou 5-9 si leur distance devient
                       inférieure à 2 mm */
        for (i = 3; i <= 4; i++)
        {
            if (0.02 > dmin2segs(OAi+i*3, OBi_o+i*3, OAi+3*(i+4), OBi_o+3*(i+4)))
            {

                sprintf(texte,"    ATTENTION CONTACT ENTRE CABLES %d ET %d",i+1, i+5);
                ibas -= 15 ; wGL->renderText (0, ibas, texte, font) ;
            }
        }

        i = Statinv (T, Fxdes, Astat, tmin, tmax, tref, tpon, 1);
        if (i == -10) {ibas -= 15; wGL->renderText (0, ibas, "   TENSIONS CABLES HORS LIMITES : ECHEC STATINV", font);}
        else if (i < 0)
        {
            sprintf(texte,"  TENSIONS CABLES HORS LIMITES : SOLUTION FAUSSE %d tensions saturées",-i);
            ibas -= 15. ; wGL->renderText (0, ibas, texte, font);
        }
        else if (i > 0)
        {
            sprintf(texte,"  SOLUTION EXACTE MAIS %d tension(s) câble(s) saturée(s)",i);
            ibas -= 15. ; wGL->renderText (0, ibas, texte, font);
        }
    }
    else matmat (Fxdes, Astat, T, 6, 9, 1) ;

    sprintf(texte,"Tensions : %+7.2lf %+7.2lf %+7.2lf %+7.2lf %+7.2lf %+7.2lf %+7.2lf %+7.2lf %+7.2lf",
            T[0],T[1],T[2],T[3],T[4],T[5],T[6],T[7],T[8]);
    ihaut += 15; wGL->renderText (0, ihaut, texte, font) ;

    sprintf(texte,"Torseur cable : %+7.2lf %+7.2lf %+7.2lf %+7.2lf %+7.2lf %+7.2lf",
            Fxdes[0],Fxdes[1],Fxdes[2],Fxdes[3],Fxdes[4],Fxdes[5]);
    ihaut += 15; wGL->renderText (0, ihaut, texte, font) ;
}

/*----------------------------------------------------------------------------------*/
/* Callback traitement touches pour actions diverses                                */
/* Voir Texte Help                                                                  */
/*----------------------------------------------------------------------------------*/
bool Sacso::keyConfig(int key, bool majuscule)
{
    bool ok = FALSE;
    int maj = majuscule ? 1 : 0;
    int smaj = -1 + 2*maj;
    double DeltaA = 2.5, DeltaL = 0.05, DeltaF = 0.5, DeltaM = 0.05 ;

    switch (key)
    {
    case Qt::Key_S : /* modèle simplifié/Exact */
        FlagExact = maj;
        break ;
    case Qt::Key_0 : /* au centre */
    case Qt::Key_O : /* au centre */
        if (fModeXouF == X_SELECT)
        {
            OA[0]=-1.0;
            OA[1]=0.0;
            OA[2]=0.0;
            Carol[0]=0.0;
            Carol[1] =0.0;
            Carol[2]=0.0;
            carolNok = true;
        }
        else
        { Fxdes[0] = Fxdes[1] = Fxdes[2] = Fxdes[3] = Fxdes[4] = Fxdes[5] =0.;}
        ok = TRUE;
        break;
    case Qt::Key_X :
        if (fModeXouF == X_SELECT) OA[0] -= smaj*DeltaL ;
        else Fxdes[0] -= smaj*DeltaF ;
        ok = TRUE;
        break ;
    case Qt::Key_Y :
        if (fModeXouF == X_SELECT) OA[1] -= smaj*DeltaL ;
        else Fxdes[1] -= smaj*DeltaF ;
        ok = TRUE;
        break ;
    case Qt::Key_Z :
        if (fModeXouF == X_SELECT) OA[2] -= smaj*DeltaL ;
        else Fxdes[2] -= smaj*DeltaF ;
        ok = TRUE;
        break ;
    case Qt::Key_R : /* Roulis */
        if (fModeXouF == X_SELECT)
        {
            Carol[2] -= smaj*DeltaA ;
            if (Carol[2]   >  180.0) Carol[2]   -= 360.0;
            if (Carol[2]   < -180.0) Carol[2]   += 360.0;
            carolNok = true;
        }
        else Fxdes[3] -= smaj*DeltaM ;
        ok = TRUE;
        break ;
    case Qt::Key_T : /* Assiette */
        if (fModeXouF == X_SELECT)
        {
            Carol[1] -= smaj*DeltaA ;
            if (Carol[1] >  180.0) Carol[1] -= 360. ;
            if (Carol[1] < -180.0) Carol[1] += 360. ;
            carolNok = true;
        }
        else Fxdes[4] -= smaj*DeltaM ;
        ok = TRUE;
        break ;
    case Qt::Key_L : /* Cap */
        if (fModeXouF == X_SELECT)
        {
            Carol[0] -= smaj*DeltaA ;
            if (Carol[0]      >  360.0) Carol[0]      -= 360.0;
            if (Carol[0]      <    0.0) Carol[0]      += 360.0;
            carolNok = true;
        }
        else Fxdes[5] -= smaj*DeltaM ;
        ok = TRUE;
        break ;
    }
    if (ok) wGL->updateGL();
    return ok;
}
